# dform.fields.py
import logging
from six import with_metaclass

from django.core.exceptions import ValidationError
from django.forms import fields
from django.forms import widgets

logger = logging.getLogger(__name__)

FIELDS = []

# ============================================================================

class MetaField(type):
    def __new__(cls, name, bases, classdict):
        global FIELDS
        klass = type.__new__(cls, name, bases, dict(classdict))
        if klass.__name__ not in ['Field', 'ChoiceField']:
            FIELDS.append(klass)
        return klass


class Field(with_metaclass(MetaField)):
    django_widget = ''
    form_control = True

    @classmethod
    def check_field_parms(cls, field_parms):
        if field_parms:
            raise ValidationError('%s field does not take parameters' % (
                cls.__name__))


class ChoiceField(Field):
    @classmethod
    def check_field_parms(cls, field_parms):
        if len(field_parms) < 1 or not isinstance(field_parms, dict):
            raise ValidationError(('%s field expected a dict of valid '
                'choices') % cls.__name__)

# ============================================================================
# Storage Types
# ============================================================================

class TextStorage(object):
    storage_key = 'answer_text'

    @classmethod
    def check_value(cls, field_parms, value):
        pass


class ChoicesStorage(object):
    storage_key = 'answer_key'

    @classmethod
    def check_value(cls, field_parms, value):
        if value not in field_parms:
            raise ValidationError('value was not in available choices')


class MultipleChoicesStorage(object):
    storage_key = 'answer_key'

    @classmethod
    def check_value(cls, field_parms, value):
        keys = value.split(',')
        for key in keys:
            if key not in field_parms:
                raise ValidationError('value was not in available choices')


class IntegerStorage(object):
    storage_key = 'answer_int'

    @classmethod
    def check_value(cls, field_parms, value):
        try:
            int(value)
        except ValueError:
            raise ValidationError('value was not an integer')


class FloatStorage(object):
    storage_key = 'answer_float'

    @classmethod
    def check_value(cls, field_parms, value):
        try:
            float(value)
        except ValueError:
            raise ValidationError('value was not a float')

# ============================================================================
# Field Types
# ============================================================================

class Text(Field, TextStorage):
    field_key = 'tx'
    django_field = fields.CharField


class MultiText(Field, TextStorage):
    field_key = 'mt'
    django_field = fields.CharField
    django_widget = widgets.Textarea


class Dropdown(ChoiceField, ChoicesStorage):
    field_key = 'dr'
    django_field = fields.ChoiceField
    django_widget = widgets.Select


class Radio(ChoiceField, ChoicesStorage):
    field_key = 'rd'
    django_field = fields.ChoiceField
    django_widget = widgets.RadioSelect
    form_control = False


class Checkboxes(ChoiceField, MultipleChoicesStorage):
    field_key = 'ch'
    django_field = fields.MultipleChoiceField
    django_widget = widgets.CheckboxSelectMultiple
    form_control = False


class Rating(Field, IntegerStorage):
    field_key = 'rt'
    django_field = fields.ChoiceField
    django_widget = widgets.RadioSelect


class Integer(Field, IntegerStorage):
    field_key = 'in'
    django_field = fields.IntegerField


class Float(Field, FloatStorage):
    field_key = 'fl'
    django_field = fields.FloatField

# ============================================================================

FIELDS_DICT = {f.field_key:f for f in FIELDS}
FIELD_CHOICES = [(f.field_key, f.__name__) for f in FIELDS]
FIELD_CHOICES_DICT = dict(FIELD_CHOICES)



from django.conf.urls import patterns, url

urlpatterns = patterns('dform.views',
    url(r'sample_survey/(\d+)/$', 'sample_survey', name='dform-sample-survey'),
    url(r'survey/(\d+)/$', 'survey', name='dform-survey'),
    url(r'survey_with_answers/(\d+)/(\d+)/$', 'survey_with_answers', 
        name='dform-survey-with-answers'),
)



import json, copy
from collections import OrderedDict
from django.core.exceptions import ValidationError
from django.core.urlresolvers import reverse, NoReverseMatch
from django.test import TestCase
from mock import Mock, patch

from awl.utils import refetch
from awl.waelsteng import AdminToolsMixin
from wrench.utils import parse_link

from dform.admin import (SurveyAdmin, SurveyVersionAdmin, QuestionAdmin,
    QuestionOrderAdmin, AnswerAdmin, AnswerGroupAdmin)
from dform.models import (Survey, SurveyVersion, EditNotAllowedException, 
    Question, QuestionOrder, Answer, AnswerGroup)
from dform.fields import (Text, MultiText, Dropdown, Radio, Checkboxes,
    Rating, Integer, Float)
from dform.forms import SurveyForm

# ============================================================================

#def pprint(data):
#    print(json.dumps(data, sort_keys=True, indent=4, separators=(',', ': ')))

class GotHere(Exception):
    pass


def perm_hook(name, *args, **kwargs):
    if name in ('survey', 'sample_survey'):
        if len(args) == 2 and len(kwargs) == 0 and args[1] == '1':
            raise GotHere()
    elif name == 'survey_with_answers':
        if len(args) == 3 and len(kwargs) == 0 and args[1] == '1' \
                and args[2] == '2':
            raise GotHere()


def sample_survey():
    # Creates and returns a survey and its questions
    survey = Survey.objects.create(name='survey')

    # add some questions
    mt = survey.add_question(MultiText, 'multi')
    tx = survey.add_question(Text, 'text value and stuff and things')
    o = OrderedDict([('a', 'Apple'), ('b', 'Bear')])
    dr = survey.add_question(Dropdown, 'drop', field_parms=o)
    o = OrderedDict([('c', 'Chair'), ('d','Dog')])
    rd = survey.add_question(Radio, 'radio', field_parms=o)
    o = OrderedDict([('e', 'Egg'), ('f', 'Fan')])
    cb = survey.add_question(Checkboxes, 'check', field_parms=o) 
    rt = survey.add_question(Rating, 'rating')
    in_ = survey.add_question(Integer, 'integer')
    fl = survey.add_question(Float, 'float')

    return survey, mt, tx, dr, rd, cb, rt, in_, fl

# ============================================================================
# Test Objects
# ============================================================================

class SurveyTests(TestCase):
    def test_survey(self):
        survey, mt, tx, dr, rd, cb, rt, in_, fl = sample_survey()

        # verify version created and correct value
        first_version = survey.latest_version
        self.assertEqual(first_version.version_num, 1)

        # verify order works
        expected_q1 = [mt, tx, dr, rd, cb, rt, in_, fl]
        for index, question in enumerate(survey.questions()):
            self.assertEqual(question, expected_q1[index])

        # -- check validation on fields
        with self.assertRaises(ValidationError):
            survey.add_question(Text, 'text', field_parms={'a':'a'})
        with self.assertRaises(ValidationError):
            survey.add_question(Dropdown, 'text')
        with self.assertRaises(ValidationError):
            survey.add_question(Dropdown, 'text', field_parms=[])

        # -- spawn a new version
        second_version = survey.new_version()
        self.assertEqual(second_version.version_num, 2)
        self.assertEqual(second_version, survey.latest_version)

        # verify order on new version
        for index, question in enumerate(survey.questions()):
            self.assertEqual(question, expected_q1[index])

        # change second version
        survey.remove_question(rt)
        tx2 = survey.add_question(Text, 'text 2', rank=1)

        # verify old and new versions are still have correct questions
        for index, question in enumerate(first_version.questions()):
            self.assertEqual(question, expected_q1[index])

        expected_q2 = [tx2, mt, tx, dr, rd, cb, in_, fl]
        for index, question in enumerate(survey.questions()):
            self.assertEqual(question, expected_q2[index])

        # -- check version is_editable
        self.assertTrue(first_version.is_editable())

        # -- test answers
        answer = 'answer and stuff and things'
        answer_group = AnswerGroup.objects.create(survey_version=first_version)
        a1 = first_version.answer_question(tx, answer_group, answer)
        self.assertEqual(a1.value, answer)

        # answer same question again, ensure replacement
        a2 = first_version.answer_question(tx, answer_group, 'something else')
        self.assertEqual(a1.id, a2.id)
        self.assertEqual(a2.value, 'something else')

        # re-check version is_editable now that it shouldn't be
        self.assertFalse(first_version.is_editable())

        # survey has answers now, shouldn't be able to edit
        with self.assertRaises(EditNotAllowedException):
            first_version.add_question(Text, 'fail')

        with self.assertRaises(EditNotAllowedException):
            first_version.remove_question(tx)

        # more answers
        a = first_version.answer_question(mt, answer_group, 'answer\nanswer')
        self.assertEqual(a.value, 'answer\nanswer')

        a = first_version.answer_question(dr, answer_group, 'a')
        self.assertEqual(a.value, 'a')
        with self.assertRaises(ValidationError):
            first_version.answer_question(dr, answer_group, 'z')

        a = first_version.answer_question(rd, answer_group, 'c')
        self.assertEqual(a.value, 'c')
        with self.assertRaises(ValidationError):
            first_version.answer_question(dr, answer_group, 'z')

        a = first_version.answer_question(cb, answer_group, 'e')
        self.assertEqual(a.value, 'e')

        a = first_version.answer_question(in_, answer_group, 1)
        self.assertEqual(a.value, 1)
        with self.assertRaises(ValidationError):
            first_version.answer_question(in_, answer_group, 'z')

        a = first_version.answer_question(fl, answer_group, 1.2)
        self.assertEqual(a.value, 1.2)
        with self.assertRaises(ValidationError):
            first_version.answer_question(fl, answer_group, 'z')

        ag2 = AnswerGroup.objects.create(survey_version=first_version)
        ag3 = AnswerGroup.objects.create(survey_version=first_version)
        a = first_version.answer_question(cb, ag2, 'e,f')
        self.assertEqual(a.value, 'e,f')
        with self.assertRaises(ValidationError):
            first_version.answer_question(cb, ag3, 'z')
        with self.assertRaises(ValidationError):
            first_version.answer_question(cb, ag3, 'e,z')

        a = first_version.answer_question(rt, answer_group, 1)
        self.assertEqual(a.value, 1)

        # check number validation
        with self.assertRaises(ValidationError):
            first_version.answer_question(rt, answer_group, 'a')

        # try to answer a question not in this version
        with self.assertRaises(AttributeError):
            survey.answer_question(rt, answer_group, 1)

        # -- misc coverage items
        str(survey)
        str(first_version)
        str(tx)
        str(QuestionOrder.objects.first())
        str(answer_group)
        str(a1)


    def test_survey_dicts(self):
        self.maxDiff = None
        survey, mt, tx, dr, rd, cb, rt, in_, fl = sample_survey()

        # -- create a delta that changes nothing
        expected = survey.to_dict()
        survey.replace_from_dict(expected)

        # verify that everything in the survey is the same, nothing new
        # created and the questions are in order
        self.assertEqual(1, Survey.objects.count())
        self.assertEqual(8, Question.objects.count())

        delta = survey.to_dict()
        self.assertEqual(expected, delta)

        # -- rename survey, change questions: reorder, edit and add
        expected['name'] = 'Renamed'
        expected['questions'][4], expected['questions'][5] = \
            expected['questions'][5], expected['questions'][4]
        expected['questions'][0].update({
            'text':'edit text label',
            'required':True,
        })
        expected['questions'].insert(0, {
            'id':0,
            'field_key':Dropdown.field_key,
            'text':'a new question',
            'required':True,
            'field_parms':OrderedDict([('g', 'Good'), ('h', 'Hello')])
        })

        # perform delta, change id of new question in "expected"
        survey.replace_from_dict(expected)

        expected['questions'][0]['id'] = Question.objects.last().id

        delta = survey.to_dict()
        self.assertEqual(expected, delta)

        # -- verify remove works
        delta = {
            'remove':[expected['questions'][0]['id'], ]
        }
        survey.replace_from_dict(delta)

        expected['questions'] = expected['questions'][1:]
        delta = survey.to_dict()
        self.assertEqual(expected, delta)

        # -- verify error conditions for bad question ids
        last_question = Question.objects.all().order_by('id').last()
        delta = {
            'questions':[{
                'id':last_question.id + 10,
            }]
        }

        with self.assertRaises(Question.DoesNotExist):
            survey.replace_from_dict(delta)

        # try remove
        delta = {
            'remove':[last_question.id + 10, ],
        }

        with self.assertRaises(Question.DoesNotExist):
            survey.replace_from_dict(delta)

        # -- verify disallowed editing
        ag = AnswerGroup.objects.create(survey_version=survey.latest_version)
        survey.answer_question(mt, ag, 'answer\nanswer')

        with self.assertRaises(EditNotAllowedException):
            survey.replace_from_dict(delta)

    def test_on_success(self):
        survey = Survey.objects.create(name='test')
        version = survey.latest_version

        with self.assertRaises(AttributeError):
            version.on_success()

        with self.settings(DFORM_SUCCESS_REDIRECT='/one/'):
            self.assertEqual('/one/', version.on_success())

            survey.success_redirect = '/two/'
            survey.save()
            survey = refetch(survey)
            version = refetch(version)
            self.assertEqual('/two/', version.on_success())

            version.success_redirect = '/three/'
            self.assertEqual('/three/', version.on_success())


def fake_reverse(name, args):
    if name in ['dform-survey', 'dform-sample-survey', 
            'dform-survey-with-answers']:
        raise NoReverseMatch

    return reverse(name, args=args)


class AdminTestCase(TestCase, AdminToolsMixin):
    def test_without_urls(self):
        # what actions are available depend on whether the dform.urls.py file
        # is used or not, and need to be able to change if it isn't there; it
        # is always there when testing, so we'll need to use mock to force the
        # code to execute
        self.initiate()
        survey_admin = SurveyAdmin(Survey, self.site)
        version_admin = SurveyVersionAdmin(SurveyVersion, self.site)
        group_admin = AnswerGroupAdmin(AnswerGroup, self.site)
        survey, _, tx, _, _, _, _, _, _ = sample_survey()

        with patch('dform.admin.reverse') as mock_reverse:
            mock_reverse.side_effect = fake_reverse

            html = self.field_value(survey_admin, survey, 'show_actions')
            actions = html.split(',')
            self.assertEqual(1, len(actions))

            url, text = parse_link(html)
            self.assertEqual(text, 'Edit Survey')

            # -- SurveyVersion of same test
            result = self.field_value(version_admin, survey.latest_version, 
                'show_actions')
            self.assertEqual(1, len(actions))

            url, text = parse_link(html)
            self.assertEqual(text, 'Edit Survey')

            # -- repeat with locked survey
            answer = 'answer and stuff and things'
            ag = AnswerGroup.objects.create(
                survey_version=survey.latest_version)
            survey.answer_question(tx, ag, answer)

            html = self.field_value(survey_admin, survey, 'show_actions')
            actions = html.split(',')
            self.assertEqual(1, len(actions))

            url, text = parse_link(html)
            self.assertEqual(text, 'New Version')

            # -- SurveyVersion of same test
            result = self.field_value(version_admin, survey.latest_version, 
                'show_actions')
            self.assertEqual('', result)

            # -- check actions on AG
            result = self.field_value(group_admin, ag, 'show_actions')
            self.assertEqual('', result)

    def assert_edit_link(self, link):
        url, text = parse_link(link)
        self.assertEqual('Edit Survey', text)

        response = self.authed_get(url)
        self.assertTemplateUsed(response, 'dform/edit_survey.html')

    def assert_sample_link(self, link):
        url, text = parse_link(link)
        self.assertEqual('View Sample', text)

        response = self.authed_get(url)
        self.assertTemplateUsed(response, 'dform/survey.html')

    def assert_change_link(self, link):
        url, text = parse_link(link)
        self.assertEqual('Answer Survey', text)

        response = self.authed_get(url)
        self.assertTemplateUsed(response, 'dform/survey.html')

    def assert_version_link(self, link):
        url, text = parse_link(link)
        self.assertEqual('New Version', text)

        response = self.authed_get(url, response_code=302)

    def test_survey_admin(self):
        self.initiate()

        survey_admin = SurveyAdmin(Survey, self.site)
        version_admin = SurveyVersionAdmin(SurveyVersion, self.site)
        survey, _, tx, _, _, _, _, _, _ = sample_survey()
        first_version = survey.latest_version

        # check version number
        result = self.field_value(survey_admin, survey, 'version_num')
        self.assertEqual('1', result)

        # -- show_actions
        links = self.field_value(survey_admin, survey, 'show_actions')
        edit, sample, change = links.split(',')
        self.assert_edit_link(edit)
        self.assert_sample_link(sample)
        self.assert_change_link(change)

        links = self.field_value(version_admin, first_version, 'show_actions')
        edit, sample, change = links.split(',')
        self.assert_edit_link(edit)
        self.assert_sample_link(sample)
        self.assert_change_link(change)

        # add answer and verify "new version" link
        answer = 'answer and stuff and things'
        ag = AnswerGroup.objects.create(survey_version=survey.latest_version)
        survey.answer_question(tx, ag, answer)

        links = self.field_value(survey_admin, survey, 'show_actions')
        version, sample, change = links.split(',')
        self.assert_version_link(version)
        self.assert_sample_link(sample)
        self.assert_change_link(change)

        links = self.field_value(version_admin, first_version, 'show_actions')
        sample, change = links.split(',')
        self.assert_sample_link(sample)
        self.assert_change_link(change)

        # -- show_versions
        self.visit_admin_link(survey_admin, survey, 'show_versions')

    def _assert_question_fields(self, html, q_url, q_text, r_url):
        q_link, r_link = html.split('|')

        url, text = parse_link(q_link)
        self.assertIn(q_url, url)
        self.assertIn(q_text, text)

        url, text = parse_link(r_link)
        self.assertIn(r_url, url)
        self.assertIn('Reorder', text)

    def _assert_link(self, html, url, text):
        u, t = parse_link(html)
        self.assertIn(url, u)
        self.assertIn(text, t)

    def test_survey_admin_links(self):
        # different test cases from test_survey_admin that need non-standard
        # surveys to be built while we're testing, so this is going in its own
        # method
        self.initiate()

        survey_admin = SurveyAdmin(Survey, self.site)
        version_admin = SurveyVersionAdmin(SurveyVersion, self.site)
        group_admin = AnswerGroupAdmin(AnswerGroup, self.site)
        survey = Survey.objects.create(name='survey')
        first_version = survey.latest_version

        # -- show_questions
        html = self.field_value(survey_admin, survey, 'show_questions')
        self.assertEqual('', html)

        html = self.field_value(version_admin, first_version, 'show_questions')
        self.assertEqual('', html)

        # add question, try again
        q_url = '/admin/dform/question/?survey_versions__id=%s' % (
            survey.latest_version.id)
        r_url = '/admin/dform/questionorder/?survey_version__id=%s' % (
            survey.latest_version.id)

        q = survey.add_question(Text, 'first question')

        html = self.field_value(survey_admin, survey, 'show_questions')
        self._assert_question_fields(html, q_url, '1 Question', r_url)
        self.assertNotIn('Questions', html)

        html = self.field_value(version_admin, first_version, 'show_questions')
        self._assert_question_fields(html, q_url, '1 Question', r_url)
        self.assertNotIn('Questions', html)

        # add question, handling multiples
        survey.add_question(Text, 'second question')

        html = self.field_value(survey_admin, survey, 'show_questions')
        self._assert_question_fields(html, q_url, '2 Questions', r_url)

        html = self.field_value(version_admin, first_version, 'show_questions')
        self._assert_question_fields(html, q_url, '2 Questions', r_url)

        # -- show answers
        url = '/admin/dform/answer/?survey_version__id=%s' % (
            survey.latest_version.id)

        # add answer
        ag = AnswerGroup.objects.create(survey_version=survey.latest_version)
        survey.answer_question(q, ag, '1st Answer')

        ch_url = '/dform/survey_with_answers/%s/%s/' % (
            survey.latest_version.id, ag.id)

        html = self.field_value(group_admin, ag, 'show_actions')
        self._assert_link(html, ch_url, 'Change Answers')

        html = self.field_value(survey_admin, survey, 'show_answers')
        self._assert_link(html, url, '1 Answer')
        self.assertNotIn('Answers', html)

        html = self.field_value(version_admin, first_version, 'show_answers')
        self._assert_link(html, url, '1 Answer')
        self.assertNotIn('Answers', html)

        # add another answer, handling multiples
        ag2 = AnswerGroup.objects.create(survey_version=survey.latest_version)
        survey.answer_question(q, ag2, '2nd Answer')

        html = self.field_value(survey_admin, survey, 'show_answers')
        self._assert_link(html, url, '2 Answers')

        html = self.field_value(version_admin, first_version, 'show_answers')
        self._assert_link(html, url, '2 Answers')

    def test_question_admin(self):
        self.initiate()

        question_admin = QuestionAdmin(Question, self.site)
        survey = Survey.objects.create(name='survey')
        version = survey.latest_version

        q1 = survey.add_question(Text, '1st question')
        q2 = survey.add_question(Text, '2nd question')

        # -- show_reorder
        url = '/admin/dform/questionorder/?survey_version__id=%s' % version.id
        html = self.field_value(question_admin, q1, 'show_reorder')
        self._assert_link(html, url, 'Reorder')

        # -- show_answers
        url = '/admin/dform/answer/?question__id=%s' % q1.id

        html = self.field_value(question_admin, q1, 'show_answers')
        self.assertEqual('', html)

        # add an answer
        ag = AnswerGroup.objects.create(survey_version=survey.latest_version)
        a1 = survey.answer_question(q1, ag, '1st Answer')
        html = self.field_value(question_admin, q1, 'show_answers')
        self._assert_link(html, url, '1 Answer')
        self.assertNotIn('Answers', html)

        # another answer
        ag2 = AnswerGroup.objects.create(survey_version=survey.latest_version)
        survey.answer_question(q1, ag2, '2nd Answer')
        html = self.field_value(question_admin, q1, 'show_answers')
        self._assert_link(html, url, '2 Answers')

        # -- question order
        qo_admin = QuestionOrderAdmin(QuestionOrder, self.site)
        qo2 = QuestionOrder.objects.last()
        text = self.field_value(qo_admin, qo2, 'show_text')
        self.assertEqual('2nd question', text)

        # move_up & move_down tested in awl.rankedmodel, just trip them for
        # coverage
        self.field_value(qo_admin, qo2, 'move_up')
        self.field_value(qo_admin, qo2, 'move_down')

        # -- answer admin
        answer_admin = AnswerAdmin(Answer, self.site)
        text = self.field_value(answer_admin, a1, 'show_text')
        self.assertEqual('1st question', text)

        text = self.field_value(answer_admin, a1, 'show_field_key')
        self.assertEqual('Text', text)

    def test_answer(self):
        self.initiate()

        answer_admin = AnswerAdmin(Answer, self.site)
        answer_group_admin = AnswerGroupAdmin(AnswerGroup, self.site)

        survey = Survey.objects.create(name='survey')
        q1 = survey.add_question(Text, '1st question')

        # create a basic AnswerGroup and test its results
        ag1 = AnswerGroup.objects.create(survey_version=survey.latest_version)

        result = self.field_value(answer_group_admin, ag1, 'show_data')
        self.assertEqual('', result)
        html = self.field_value(answer_group_admin, ag1, 'show_questions')
        url, text = parse_link(html)
        self.assertEqual('1 Question', text)
        result = self.field_value(answer_group_admin, ag1, 'show_answers')
        self.assertEqual('', result)

        # add a question and check for change
        q2 = survey.add_question(Text, '2nd question')

        html = self.field_value(answer_group_admin, ag1, 'show_questions')
        url, text = parse_link(html)
        self.assertEqual('2 Questions', text)

        # add an answer and check for change
        answer = survey.answer_question(q1, ag1, 'stuff')
        html = self.field_value(answer_group_admin, ag1, 'show_answers')
        url, text = parse_link(html)
        self.assertEqual('1 Answer', text)

        # add another answer
        answer = survey.answer_question(q2, ag1, 'stuff')
        html = self.field_value(answer_group_admin, ag1, 'show_answers')
        url, text = parse_link(html)
        self.assertEqual('2 Answers', text)


# ============================================================================
# Test Views
# ============================================================================

class SurveyAdminViewTests(TestCase, AdminToolsMixin):
    def test_survey_delta_view(self):
        self.initiate()
        self.maxDiff = None
        survey, mt, tx, dr, rd, cb, rt, in_, fl = sample_survey()

        # -- create a new survey
        expected = {
            'name':'New Survey',
            'questions':[{
                'id':0,
                'field_key':Dropdown.field_key,
                'text':'a new question',
                'required':True,
                'field_parms':OrderedDict([('g', 'Good'), ('h', 'Hello')])
            }],
        }

        response = self.authed_post('/dform_admin/survey_delta/0/',
            data={'delta':json.dumps(expected)})

        expected['questions'][0]['id'] = Question.objects.last().id

        survey2 = Survey.objects.last()
        delta = survey2.to_dict()
        self.assertEqual(expected, delta)

        self.assertEqual(2, Survey.objects.count())
        self.assertEqual(9, Question.objects.count())

        # -- verify mismatched question/survey is refused
        delta = {
            'questions':[{
                'id':mt.id,   # question from survey1
            }]
        }

        response = self.authed_post(
            '/dform_admin/survey_delta/%s/' % survey2.latest_version.id, 
            response_code=404, data={'delta':json.dumps(delta)})

        # -- verify error conditions for bad question ids
        last_question = Question.objects.all().order_by('id').last()
        delta = {
            'questions':[{
                'id':last_question.id + 10,
            }]
        }

        response = self.authed_post(
            '/dform_admin/survey_delta/%s/' % survey.latest_version.id, 
            response_code=404, data={'delta':json.dumps(delta)})

        # try remove
        delta = {
            'remove':[last_question.id + 10, ],
        }

        response = self.authed_post(
            '/dform_admin/survey_delta/%s/' % survey.latest_version.id, 
            response_code=404, data={'delta':json.dumps(delta)})

        # -- verify disallowed editing
        ag = AnswerGroup.objects.create(survey_version=survey.latest_version)
        survey.answer_question(mt, ag, 'answer\nanswer')
        delta = {
            'name':'Foo',
        }

        response = self.authed_post(
            '/dform_admin/survey_delta/%s/' % survey.latest_version.id, 
            response_code=404, data={'delta':json.dumps(delta)})

    def test_survey_edit(self):
        # Only does rudimentary execution tests, doesn't test the actual 
        # javascript editor
        self.initiate()

        response = self.authed_get('/dform_admin/survey_editor/0/')
        self.assertTemplateUsed(response, 'dform/edit_survey.html')

        survey = Survey.objects.first()

        response = self.authed_get('/dform_admin/survey_editor/%s/' %
            survey.latest_version.id)
        self.assertTemplateUsed(response, 'dform/edit_survey.html')


class SurveyViewTests(TestCase):
    def test_sample_survey(self):
        survey, mt, tx, dr, rd, cb, rt, in_, fl = sample_survey()

        response = self.client.get('/dform/sample_survey/%s/' % (
            survey.latest_version.id))
        self.assertEqual(200, response.status_code)
        self.assertEqual('Sample: survey', response.context['title'])
        self.assertTemplateUsed(response, 'dform/survey.html')

    def test_survey_view(self):
        redirect = '/admin/'
        survey, mt, tx, dr, rd, cb, rt, in_, fl = sample_survey()
        survey.success_redirect = redirect
        survey.save()

        response = self.client.get('/dform/survey/%s/' % (
            survey.latest_version.id))
        self.assertEqual(200, response.status_code)
        self.assertEqual('survey', response.context['title'])
        self.assertTemplateUsed(response, 'dform/survey.html')

        dataset = [(mt, 'mt'), (tx, 'tx'), (dr, 'a'), (rd, 'c'), (cb, 'e'), 
            (rt, 2), (in_, 42), (fl, 13.69)]

        data = {}
        for question, value in dataset:
            key = 'q_%s' % question.id
            data[key] = value

        response = self.client.post('/dform/survey/%s/' % (
            survey.latest_version.id), data)
        self.assertEqual(302, response.status_code)
        self.assertIn(redirect, response._headers['location'][1])

        self.assertEqual(8, Answer.objects.count())
        for question, value in dataset:
            answer = Answer.objects.get(question=question)
            self.assertEqual(value, answer.value)

        # -- check with system level submit action
        with self.settings(DFORM_SURVEY_SUBMIT='/foo/'):
            response = self.client.get('/dform/survey/%s/' % (
                survey.latest_version.id))
            self.assertEqual('/foo/', response.context['submit_action'])

    def test_permission_hooks(self):
        hook = 'dform.tests.test_dform.perm_hook'

        with self.settings(DFORM_PERMISSION_HOOK=hook):
            with self.assertRaises(GotHere):
                self.client.post('/dform/survey/1/')

            with self.assertRaises(GotHere):
                self.client.post('/dform/sample_survey/1/')

            with self.assertRaises(GotHere):
                self.client.post('/dform/survey_with_answers/1/2/')

    def test_survey_with_answers(self):
        redirect = '/admin/'
        survey, mt, tx, dr, rd, cb, rt, in_, fl = sample_survey()
        survey.success_redirect = redirect
        survey.save()

        dataset = [(mt, 'mt'), (tx, 'tx'), (dr, 'a'), (rd, 'c'), (cb, 'e,f'), 
            (rt, 2), (in_, 42), (fl, 13.69)]

        ag = AnswerGroup.objects.create(survey_version=survey.latest_version)
        for question, value in dataset:
            survey.answer_question(question, ag, value)

        response = self.client.get('/dform/survey_with_answers/%s/%s/' % (
            survey.latest_version.id, ag.id))
        self.assertEqual(200, response.status_code)
        self.assertEqual('survey', response.context['title'])
        self.assertTemplateUsed(response, 'dform/survey.html')

        form = response.context['form']
        self.assertEqual(8, len(form.fields))
        for question, value in dataset:
            key = 'q_%s' % question.id
            field = form.fields[key]
            self.assertEqual(value, field.initial)

        # -- test change in values, including emptying out a field (no cb)
        dataset = [(mt, 'mt2'), (tx, 'tx2'), (dr, 'b'), (rd, 'd'), (rt, 3), 
            (in_, 43), (fl, 18.62)]
        data = {}
        for question, value in dataset:
            key = 'q_%s' % question.id
            data[key] = value

        response = self.client.post('/dform/survey_with_answers/%s/%s/' % (
            survey.latest_version.id, ag.id), data)
        self.assertEqual(302, response.status_code)
        self.assertIn(redirect, response._headers['location'][1])

        self.assertEqual(7, Answer.objects.count())
        for question, value in dataset:
            answer = Answer.objects.get(question=question)
            self.assertEqual(value, answer.value)

        # once more with a lot of missing fields to trigger the answer delete
        # handling
        data = {
            'q_%s' % mt.id:'mt3',
        }

        response = self.client.post('/dform/survey_with_answers/%s/%s/' % (
            survey.latest_version.id, ag.id), data)
        self.assertEqual(302, response.status_code)
        self.assertIn(redirect, response._headers['location'][1])

        self.assertEqual(1, Answer.objects.count())
        answer = Answer.objects.get(question=mt)
        self.assertEqual('mt3', answer.value)

        # -- check with system level submit action
        with self.settings(DFORM_SURVEY_WITH_ANSWERS_SUBMIT='/foo/'):
            response = self.client.get('/dform/survey_with_answers/%s/%s/' % (
                survey.latest_version.id, ag.id))
            self.assertEqual('/foo/', response.context['submit_action'])


class FormTest(TestCase):
    def test_form(self):
        # coverage for edge cases not covered by SurveyViewTests

        with self.assertRaises(AttributeError):
            SurveyForm(initial={'foo':'bar'}, survey_version=None)

        survey = Survey.objects.create(name='survey', success_redirect='/foo/')
        survey.add_question(Text, '1st question')
        form = SurveyForm(survey_version=survey.latest_version)

        self.assertFalse(form.has_required())

        survey.add_question(Text, '2nd question', required=True)
        form = SurveyForm(survey_version=survey.latest_version)

        self.assertTrue(form.has_required())



# dform.forms.py
from django import forms
from django.template.loader import render_to_string

from .fields import (FIELDS_DICT, ChoiceField, Rating, MultipleChoicesStorage,
    Integer, Float)
from .models import Answer, AnswerGroup, Question

# ============================================================================

class SurveyForm(forms.Form):
    def __init__(self, *args, **kwargs):
        self.survey_version = kwargs.pop('survey_version')
        self.answer_group = kwargs.pop('answer_group', None)

        if 'initial' in kwargs:
            raise AttributeError(
                '"initial" keyword is not allowed with SurveyForm')

        super(SurveyForm, self).__init__(*args, **kwargs)

        # populate any answers from the database
        values = {}
        if self.answer_group:
            for answer in self.answer_group.answer_set.all():
                key = 'q_%s' % answer.question.id
                values[key] = answer.value

        # update values with info from a POST if passed in
        if len(args) > 0:
            values.update(args[0])

        self.populate_fields(values)

    def populate_fields(self, values):
        for question in self.survey_version.questions():
            name = 'q_%s' % question.id

            kwargs = {
                'label':question.text,
                'required':question.required,
            }

            if name in values:
                kwargs['initial'] = values[name]

            if question.field.django_widget:
                kwargs['widget'] = question.field.django_widget

            if question.field == Rating:
                kwargs['choices'] = (
                    (5, '5 Star'),
                    (4, '4 Star'),
                    (3, '3 Star'),
                    (2, '2 Star'),
                    (1, '1 Star'),
                )
            elif issubclass(question.field, ChoiceField):
                kwargs['choices'] = question.field_choices()

            field = question.field.django_field(**kwargs)
            field.question = question
            if question.field.form_control:
                field.widget.attrs['class'] = 'form-control'

            self.fields[name] = field

    def render_form(self):
        return render_to_string('dform/fields.html', {'form':self})

    def save(self):
        if not self.answer_group:
            self.answer_group = AnswerGroup.objects.create(
                survey_version=self.survey_version)

        for name, field in self.fields.items():
            question = Question.objects.get(id=name[2:], 
                survey_versions=self.survey_version)

            value = self.cleaned_data[name]
            if not value:
                # value is empty, remove any existing answers and otherwise
                # ignore it
                try:
                    # check for an existing Answer that should now be removed
                    answer = Answer.objects.get(question=question,
                        answer_group=self.answer_group)
                    answer.delete()
                except Answer.DoesNotExist:
                    # no answer to remove, do nothing
                    pass
            else:
                if question.field in [Rating, Integer]:
                    value = int(value)
                elif question.field == Float:
                    value = float(value)
                elif issubclass(question.field, MultipleChoicesStorage):
                    value = ','.join(value)

                self.survey_version.answer_question(question, 
                    self.answer_group, value)

    def has_required(self):
        for field in self.fields.values():
            if field.required:
                return True

        return False



from django.conf.urls import patterns, url

urlpatterns = patterns('dform.views',
    url(r'survey_delta/(\d+)/$', 'survey_delta', name='dform-survey-delta'),
    url(r'survey_editor/(\d+)/$', 'survey_editor', name='dform-edit-survey'),
    url(r'new_version/(\d+)/$', 'new_version', name='dform-new-version'),
)



from django.contrib import admin
from django.core.urlresolvers import reverse, NoReverseMatch

from awl.admintools import make_admin_obj_mixin
from awl.rankedmodel.admintools import admin_link_move_up, admin_link_move_down

from .fields import FIELD_CHOICES_DICT
from .models import (Survey, SurveyVersion, Question, QuestionOrder, Answer,
    AnswerGroup)

# ============================================================================

def _questions_link(version, show_reorder=True):
    num_q = Question.objects.filter(survey_versions=version).count()
    if num_q == 0:
        return ''

    plural = ''
    if num_q > 1:
        plural = 's'

    show = reverse('admin:dform_question_changelist')
    reorder = reverse('admin:dform_questionorder_changelist')

    urls = [
        '<a href="%s?survey_versions__id=%s">%s Question%s</a>' % (show, 
            version.id, num_q, plural)
    ]

    if show_reorder:
        urls.append(
            '<a href="%s?survey_version__id=%s">Reorder</a>' % (reorder, 
                version.id)
        )

    return '&nbsp;|&nbsp'.join(urls)


def _answers_link(version):
    num_a = Answer.objects.filter(answer_group__survey_version=version).count()
    if num_a == 0:
        return ''

    plural = ''
    if num_a > 1:
        plural = 's'

    link = reverse('admin:dform_answer_changelist')

    url = '<a href="%s?survey_version__id=%s">%s Answer%s</a>' % (link,
        version.id, num_a, plural)
    return url

# ============================================================================
# Surveys
# ============================================================================

@admin.register(Survey)
class SurveyAdmin(admin.ModelAdmin):
    list_display = ('id', 'name', 'version_num', 'show_actions',
        'show_versions', 'show_questions', 'show_answers')

    def version_num(self, obj):
        return '%s' % obj.latest_version.version_num
    version_num.short_description = 'Latest Version'

    def show_actions(self, obj):
        actions = []

        if obj.latest_version.is_editable():
            url = reverse('dform-edit-survey', args=(obj.latest_version.id,))
            actions.append('<a href="%s">Edit Survey</a>' % url)
        else:
            url = reverse('dform-new-version', args=(obj.id,))
            actions.append('<a href="%s">New Version</a>' % url)

        try:
            url = reverse('dform-sample-survey', args=(obj.latest_version.id,))
            actions.append('<a href="%s">View Sample</a>' % url)
        except NoReverseMatch:
            # sample-survey view isn't guaranteed to be there
            pass

        try:
            url = reverse('dform-survey', args=(obj.latest_version.id,))
            actions.append('<a href="%s">Answer Survey</a>' % url)
        except NoReverseMatch:
            # survey view isn't guaranteed to be there
            pass

        return ', '.join(actions)
    show_actions.short_description = 'Actions'
    show_actions.allow_tags = True

    def show_versions(self, obj):
        num_v = SurveyVersion.objects.filter(survey=obj).count()
        link = reverse('admin:dform_surveyversion_changelist')
        url = '<a href="%s?survey__id=%s">%s Versions</a>' % (link, obj.id, 
            num_v)

        return url
    show_versions.short_description = 'Versions'
    show_versions.allow_tags = True

    def show_questions(self, obj):
        return _questions_link(obj.latest_version)
    show_questions.short_description = 'Questions'
    show_questions.allow_tags = True

    def show_answers(self, obj):
        return _answers_link(obj.latest_version)
    show_answers.short_description = 'Answers'
    show_answers.allow_tags = True


mixin = make_admin_obj_mixin('SurveyVersionMixin')
mixin.add_obj_link('show_survey', 'survey')

@admin.register(SurveyVersion)
class SurveyVersionAdmin(admin.ModelAdmin, mixin):
    list_display = ('id', 'show_survey', 'version_num', 'show_actions',
        'show_questions', 'show_answers')

    def show_actions(self, obj):
        actions = []
        if obj.is_editable():
            url = reverse('dform-edit-survey', args=(obj.id,))
            actions.append('<a href="%s">Edit Survey</a>' % url)

        try:
            url = reverse('dform-sample-survey', args=(obj.id,))
            actions.append('<a href="%s">View Sample</a>' % url)
        except NoReverseMatch:
            # view sample isn't guaranteed to be there
            pass

        try:
            url = reverse('dform-survey', args=(obj.id,))
            actions.append('<a href="%s">Answer Survey</a>' % url)
        except NoReverseMatch:
            # survey view isn't guaranteed to be there
            pass

        return ', '.join(actions)
    show_actions.short_description = 'Actions'
    show_actions.allow_tags = True

    def show_questions(self, obj):
        return _questions_link(obj)
    show_questions.short_description = 'Questions'
    show_questions.allow_tags = True

    def show_answers(self, obj):
        return _answers_link(obj)
    show_answers.short_description = 'Answers'
    show_answers.allow_tags = True

# ============================================================================
# Questions
# ============================================================================

@admin.register(Question)
class QuestionAdmin(admin.ModelAdmin):
    list_display = ('id', 'text', 'field_key', 'required', 'show_reorder',
        'show_answers')

    def show_reorder(self, obj):
        link = reverse('admin:dform_questionorder_changelist')
        url = '<a href="%s?survey_version__id=%s">Reorder</a>' % (link, 
            obj.survey.latest_version.id)

        return url
    show_reorder.short_description = 'Reorder'
    show_reorder.allow_tags = True

    def show_answers(self, obj):
        num_a = Answer.objects.filter(question=obj).count()
        if num_a == 0:
            return ''

        plural = ''
        if num_a > 1:
            plural = 's'

        link = reverse('admin:dform_answer_changelist')
        url = '<a href="%s?question__id=%s">%s Answer%s</a>'  % (link, obj.id, 
            num_a, plural)
        return url
    show_answers.short_description = 'Answers'
    show_answers.allow_tags = True


@admin.register(QuestionOrder)
class QuestionOrderAdmin(admin.ModelAdmin):
    list_display = ('id', 'survey_version', 'show_text', 'move_up',
        'move_down')

    def show_text(self, obj):
        return obj.question.text
    show_text.short_description = 'Question Text'

    def move_up(self, obj):
        return admin_link_move_up(obj, 'Up')
    move_up.allow_tags = True
    move_up.short_description = 'Move Up'

    def move_down(self, obj):
        return admin_link_move_down(obj, 'Down')
    move_down.allow_tags = True
    move_down.short_description = 'Move Down'

# ============================================================================
# Answers
# ============================================================================

mixin = make_admin_obj_mixin('AnswerMixin')
mixin.add_obj_link('show_group', 'answer_group',
    display='AnswerGroup.id={{obj.id}}')
mixin.add_obj_link('show_question', 'question',
    display='Question.id={{obj.id}}')

@admin.register(Answer)
class AnswerAdmin(admin.ModelAdmin, mixin):
    list_display = ('id', 'show_group', 'show_question', 'show_text', 
        'show_field_key', 'value')

    def show_text(self, obj):
        return obj.question.text
    show_text.short_description = 'Question Text'

    def show_field_key(self, obj):
        return FIELD_CHOICES_DICT[obj.question.field_key]
    show_field_key.short_description = 'Field Key'


mixin = make_admin_obj_mixin('AnswerGroupMixin')
mixin.add_obj_link('show_data', 'group_data')
mixin.add_obj_link('show_version', 'survey_version',
    display='SurveyVersion.id={{obj.id}}')

@admin.register(AnswerGroup)
class AnswerGroupAdmin(admin.ModelAdmin, mixin):
    list_display = ('id', 'updated', 'show_version', 'show_data', 
        'show_questions', 'show_answers', 'show_actions')

    def show_questions(self, obj):
        return _questions_link(obj.survey_version, False)
    show_questions.short_description = 'Questions'
    show_questions.allow_tags = True

    def show_answers(self, obj):
        num_a = Answer.objects.filter(answer_group=obj).count()
        if num_a == 0:
            return ''

        plural = ''
        if num_a > 1:
            plural = 's'

        link = reverse('admin:dform_answer_changelist')

        url = '<a href="%s?answer_group__id=%s">%s Answer%s</a>' % (link,
            obj.id, num_a, plural)
        return url
    show_answers.short_description = 'Answers'
    show_answers.allow_tags = True

    def show_actions(self, obj):
        try:
            url = reverse('dform-survey-with-answers', args=(
                obj.survey_version.id, obj.id))
            return '<a href="%s">Change Answers</a>' % url
        except NoReverseMatch:
            # view survey-with-answers isn't guaranteed to be there
            return ''
    show_actions.short_description = 'Actions'
    show_actions.allow_tags = True



from django.conf.urls import include, url
from django.contrib import admin

urlpatterns = [
    url(r'admin/', include(admin.site.urls)),

    url(r'rankedmodel/', include('awl.rankedmodel.urls')),

    url(r'dform/', include('dform.urls')),
    url(r'dform_admin/', include('dform.admin_urls')),
]



# dform.models.py
import logging, collections

from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models, transaction
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.template import Context, Template
from django.template.loader import render_to_string
from django.utils.encoding import python_2_unicode_compatible

from jsonfield import JSONField
from awl.models import TimeTrackModel
from awl.rankedmodel.models import RankedModel

from .fields import FIELD_CHOICES, FIELDS_DICT

logger = logging.getLogger(__name__)

# ============================================================================
# Survey Management
# ============================================================================

class EditNotAllowedException(Exception):
    """Exception thrown if an attempt is made to edit a version of a survey
    that currently has answers associated with it."""
    pass


@python_2_unicode_compatible
class Survey(TimeTrackModel):
    """Main class that encapsulates a survey.  The actual questions are
    associated with a version of the survey (:class:`SurveyVersion), this is a
    container for all of the versions that are associated together.  The most
    recent version is available in the :func:`Survey.latest_version` property.

    .. note::

        A :class:`SurveyVersion` object is created automatically via a signal
        when a new instance of this class is saved.
    """
    name = models.CharField(max_length=50)
    success_redirect = models.TextField()

    def __str__(self):
        return 'Survey(id=%s %s)' % (self.id, self.name)

    @transaction.atomic
    def new_version(self):
        """Creates a new version of the ``Survey``, associating all the
        current verion's questions with the new version.  Note, this will have
        the side effect of changing the latest version of this ``Survey``.

        :returns:
            newly created :class:`SurveyVersion`
        """
        old_version = self.latest_version
        new_version = SurveyVersion.objects.create(survey=self,
            version_num=self.latest_version.version_num + 1)

        orders = QuestionOrder.objects.filter(
            survey_version=old_version).order_by('rank')
        for order in orders:
            QuestionOrder.objects.create(survey_version=new_version,
                question=order.question, rank=order.rank)
            order.question.survey_versions.add(new_version)

        return new_version

    @property
    def latest_version(self):
        return SurveyVersion.objects.filter(survey=self).order_by(
            '-version_num')[0]

    # ---------
    # Methods on Latest Version

    def add_question(self, field, text, rank=0, required=False, 
            field_parms={}):
        """Convenience method for :func:`SurveyVersion.add_question` using
        the latest version.

        :param field: 
            a :class:`Field` instance describing the kind of question being
            created
        :param text:
            text to display when asking the question
        :param rank:
            the order number of this question, defaults to 0 which means
            insert last
        :param required:
            whether the question is required for form submission
        :param field_parms:
            a field specific dictionary specifying parameters for the use of
            the field.  
        :returns:
            newly created :class:`Question` object
        :raises EditNotAllowedException:
            editing is not allowed for surveys that already have answers
        """
        return self.latest_version.add_question(field, text, rank, required,
            field_parms)
        
    def remove_question(self, question):
        """Convenience method for :func:`SurveyVersion.remove_question` using
        the latest version.

        :param question:
            :class:`Question` to be removed
        :raises EditNotAllowedException:
            editing is not allowed for surveys that already have answers
        """
        return self.latest_version.remove_question(question)


    def questions(self):
        """Convenience method for :func:`SurveyVersion.questions` on latest
        version of survey.
        
        :returns:
            list of :class:`Question` objects
        """
        return self.latest_version.questions()

    def answer_question(self, question, answer_group, value):
        """Convenience method for :func:`SurveyVersion.answer_question` on
        latest version of survey.

        :param question:
            :class:`Question` that is being answered.  It must be a valid
            question for this version of the survey
        :param answer_group:
            A number that groups together the answers for a survey.  An
            example would be to use the user's id so that all of their answers
            are grouped together.  This would mean each user could only answer
            each survey once.
        :param value:
            Value to record as an answer.  The value is validated against the
            field associated with the :class:`Question`.

        :raises ValidationError:
            If the value given does not pass the question's field's validation
        :raises AttributeError:
            If the question is not attached to this version of the
            ``Survey``
        """
        return self.latest_version.answer_question(question, answer_group,
            value)

    def to_dict(self):
        """Convenience method for :func:`SurveyVersion.to_dict` on the latest
        version of the survey.
        """
        return self.latest_version.to_dict()

    def replace_from_dict(self, data):
        """Convenience method for :func:`SurveyVersion.replace_from_dict` on
        the latest version of the survey.

        :param data:
            Dictionary to overwrite the contents of the ``Survey`` and
            associated :class:`Question` objects with.
        :raises EditNotAllowedException:
            If the version being replaced is active
        :raises Question.DoesNotExist:
            If a question id is referenced that does not exist or is not
            associated with this survey.
        """
        return self.latest_version.replace_from_dict(data)


@receiver(post_save, sender=Survey)
def survey_post_save(sender, **kwargs):
    if kwargs['created']:
        # newly created object, create a version to go with it
        SurveyVersion.objects.create(survey=kwargs['instance'])


@python_2_unicode_compatible
class SurveyVersion(TimeTrackModel):
    """An container for questions and answers for a survey.  Multiple versions
    can be associated with a single :class:`Survey` object, allowing reporting
    over time even as questions change.
    """
    survey = models.ForeignKey(Survey)
    version_num = models.PositiveSmallIntegerField(default=1)
    success_redirect = models.TextField(blank=True)

    def __str__(self):
        return 'SurveyVersion(id=%s survey=%s, num=%s)' % (self.id, 
            self.survey.name, self.version_num)

    class Meta:
        verbose_name = 'Survey Version'

    def validate_editable(self):
        """Raises :class:`EditNotAllowedException` if there are
        :class:`Answer` objects associated with this version.

        :raises EditNotAllowedException:
        """
        if Answer.objects.filter(
                answer_group__survey_version=self).count() != 0:
            raise EditNotAllowedException()

    def is_editable(self):
        """Returns ``True`` if there are no :class:`Answer` objects associated
        with this version."""
        count = Answer.objects.filter(answer_group__survey_version=self).count()
        return count == 0

    def on_success(self):
        """Called when this survey version has been successfully submitted.
        Uses the ``success_redirect`` field, if it is empty then uses the
        associated :class:`Survey` object's ``success_redirect`` field.  The
        fields is processed as a django template with this survey version as
        context and the result is returned.
        """
        redirect = ''
        if hasattr(settings, 'DFORM_SUCCESS_REDIRECT'):
            redirect = settings.DFORM_SUCCESS_REDIRECT

        if self.survey.success_redirect:
            redirect = self.survey.success_redirect

        if self.success_redirect:
            redirect = self.success_redirect

        if not redirect:
            raise AttributeError('No successful redirection URL defined!')

        template = Template(redirect)
        context = Context({'survey_version':self})
        return template.render(context)

    def add_question(self, field, text, rank=0, required=False, field_parms={}):
        """Creates a new :class:`Question` for this ``SurveyVersion``.

        :param field: 
            a :class:`Field` instance describing the kind of question being
            created
        :param text:
            text to display when asking the question
        :param rank:
            the order number of this question, defaults to 0 which means
            insert last
        :param required:
            whether the question is required for form submission
        :param field_parms:
            a field specific dictionary specifying parameters for the use of
            the field.  
        :returns:
            newly created :class:`Question` object
        :raises EditNotAllowedException:
            editing is not allowed for surveys that already have answers
        """
        self.validate_editable()

        field.check_field_parms(field_parms)
        question = Question.objects.create(survey=self.survey, text=text,
            field_key=field.field_key, required=required,
            field_parms=field_parms)
        question.survey_versions.add(self)

        kwargs = {
            'survey_version':self,
            'question':question,
        }
        if rank != 0:
            kwargs['rank'] = rank

        QuestionOrder.objects.create(**kwargs)
        return question

    def remove_question(self, question):
        """Removes the given question from this ``SurveyVersion``. 

        :param question:
            :class:`Question` to be removed
        :raises EditNotAllowedException:
            editing is not allowed for surveys that already have answers
        """
        self.validate_editable()
        question.survey_versions.remove(self)
        QuestionOrder.objects.get(question=question, 
            survey_version=self).delete()

    def questions(self):
        """Returns an iterable of the questions for this survey version in
        order.

        :returns:
            list of :class:`Question` objects
        """
        orders = QuestionOrder.objects.filter(survey_version=self
            ).order_by('rank')
        return [order.question for order in orders]

    def answer_question(self, question, answer_group, value):
        """Record an answer to the given question in this version of the
        survey.

        :param question:
            :class:`Question` that is being answered.  It must be a valid
            question for this version of the survey
        :param answer_group:
            A number that groups together the answers for a survey.  An
            example would be to use the user's id so that all of their answers
            are grouped together.  This would mean each user could only answer
            each survey once.
        :param value:
            Value to record as an answer.  The value is validated against the
            field associated with the :class:`Question`.

        :raises ValidationError:
            If the value given does not pass the question's field's validation
        :raises AttributeError:
            If the question is not attached to this version of the
            ``Survey``
        """
        try:
            # make sure this question is registered against this survey
            # version
            question.survey_versions.get(id=self.id)
        except SurveyVersion.DoesNotExist:
            raise AttributeError()

        return Answer.factory(question, answer_group, value)

    def to_dict(self):
        """Returns a dictionary representation of this survey version.

        Format:

        .. code-block::python

            {
                'name':survey_name,
                'questions':[
                    {
                        'id':question_id,
                        'field_key':question_field_key,
                        'text':question_text,
                        'required':is_question_required,
                        'field_params:OrderedDict(*field_parm_tuples),
                    }
                ]
            }

        .. note::

            The order of the question dictionaries will be the order of the
            questions in the survey.

        .. note::

            Be careful to use a ``OrderedDict`` with the field parameters, as
            the order of parameters is used for the order of display in
            multiple choice style questions.
        """
        questions = []
        for question in self.questions():
            questions.append({
                'id':question.id,
                'field_key':question.field_key,
                'text':question.text,
                'required':question.required,
                'field_parms':question.field_parms,
            })

        data = {
            'name':self.survey.name,
            'questions':questions,
        }

        return data

    def replace_from_dict(self, data):
        """Takes the given dictionary and modifies this survey version and its
        associated questions.  Uses the same format as 
        :func:`SurveyVersion.to_dict` with one additional (optional) key
        "remove" which contains a list of :class:`Question` ids to be removed
        from the ``Survey``.

        :param data:
            Dictionary to overwrite the contents of the ``Survey`` and
            associated :class:`Question` objects with.
        :raises EditNotAllowedException:
            If the version being replaced is active
        :raises Question.DoesNotExist:
            If a question id is referenced that does not exist or is not
            associated with this survey.
        """
        self.validate_editable()

        if 'name' in data:
            self.survey.name = data['name']
            self.survey.save()

        if 'questions' in data:
            for q_data in data['questions']:
                if q_data['id'] == 0:
                    # new question
                    kwargs = {
                        'field':FIELDS_DICT[q_data['field_key']],
                        'text':q_data['text'],
                        'required':q_data['required'],
                        'field_parms':q_data['field_parms'],
                    }

                    # add the question and set the data's question id so that we
                    # can do re-ordering down below
                    question = self.add_question(**kwargs)
                    q_data['id'] = question.id
                else:
                    question = Question.objects.get(id=q_data['id'],
                        survey_versions__id=self.id)
                   
                    question.text = q_data['text']
                    question.required = q_data['required']
                    question.field_parms = q_data['field_parms']
                    question.save()

        # fix the ranking -- creating new questions will mess with the order
        if 'questions' in data:
            for index, q_data in enumerate(data['questions']):
                question = Question.objects.get(id=q_data['id'],
                    survey_versions__id=self.id)

                q_order = QuestionOrder.objects.get(question=question,
                    survey_version=self)
                q_order.rank = index + 1
                q_order.save(rerank=False)

        if 'remove' in data:
            for id in data['remove']:
                question = Question.objects.get(id=id,
                    survey_versions__id=self.id)
                self.remove_question(question)

# ============================================================================
# Question & Answers
# ============================================================================

@python_2_unicode_compatible
class Question(TimeTrackModel):
    """This class represents a question in the survey.  It may be associated
    with multiple :class:`SurveyVersion` objects that are connected to the
    same :class:`Survey`.

    The prefered way of constructing this object is to use the
    :func:`SurveyVersion.add_question` method.
    """
    survey = models.ForeignKey(Survey)
    survey_versions = models.ManyToManyField(SurveyVersion)

    text = models.TextField(blank=True)
    field_key = models.CharField(max_length=2, choices=FIELD_CHOICES)
    field_parms = JSONField(default={}, blank=True,
        load_kwargs={'object_pairs_hook':collections.OrderedDict})
    required = models.BooleanField(default=False)

    def __str__(self):
        return 'Question(id=%s %s:%s)' % (self.id, self.field_key,
            self.short_text)

    @property
    def field(self):
        """Property that returns the :class:`Field` class for this
        question."""
        return FIELDS_DICT[self.field_key]

    def field_choices(self):
        """Returns a django-style "choices" tuple set based on the parameters
        allowed for this field.
        """
        choices = []
        for k, v in self.field_parms.items():
            choices.append((k, v))

        return choices

    @property
    def short_text(self):
        """Returns a short version of the question text.  Any thing longer
        than 15 characters is truncated with an elipses."""
        text = self.text
        if len(text) >= 15:
            text = text[:12] + '...'

        return text


@python_2_unicode_compatible
class QuestionOrder(TimeTrackModel, RankedModel):
    """:class:`Question` objects need to be stored in consistent order in a
    :class:`SurveyVersion`, this class tracks their ordering within a
    :class:`SurveyVersion`.  This is treated as a separated object from the
    :class:`Question` itself so that the same questions can have different
    orders in different versions.
    """
    survey_version = models.ForeignKey(SurveyVersion)
    question = models.ForeignKey(Question)

    class Meta:
        verbose_name = 'Question Order'
        verbose_name_plural = 'Question Order'
        ordering = ['survey_version__id', 'rank']

    def __str__(self):
        return 'QuestionOrder(id=%s, rank=%s sv.id=%s q.id=%s)' % (self.id,
            self.rank, self.survey_version.id, self.question.id)

    def grouped_filter(self):
        return QuestionOrder.objects.filter(
            survey_version=self.survey_version).order_by('rank')


@python_2_unicode_compatible
class AnswerGroup(TimeTrackModel):
    """Groups together a set of :class:`Answer` objects for a single response
    to a :class:`SurveyVersion`.

    :param group_data: a :class:`GenericForeignKey` that can be used to
        associate the group of answers with some identifying piece of data,
        for example the :class:`User` class of the respondent.  Can be left
        blank.
    """
    survey_version = models.ForeignKey(SurveyVersion)

    content_type = models.ForeignKey(ContentType, null=True, blank=True)
    object_id = models.PositiveIntegerField(default=0)
    group_data = GenericForeignKey('content_type', 'object_id')

    class Meta:
        verbose_name = 'Answer Group'

    def __str__(self):
        return 'AnswerGroup(id=%s data=%s)' % (self.id, self.group_data)


@python_2_unicode_compatible
class Answer(TimeTrackModel):
    """Stores a single answer to a :class:`Question` in a survey.  Uses sparse
    storage for different data types.  

    The prefered method of creating :class:`Answer` objects is to call
    :func:`SurveyVersion.answer_question`
    """
    question = models.ForeignKey(Question)
    answer_group = models.ForeignKey(AnswerGroup)

    answer_text = models.TextField(blank=True)
    answer_key = models.TextField(blank=True)
    answer_int = models.IntegerField(null=True, blank=True)
    answer_float = models.FloatField(null=True, blank=True)

    def __str__(self):
        return 'Answer(id=%s ag.id=%s q.id=%s value=%s)' % (self.id, 
            self.answer_group.id, self.question.id, self.display_value)

    @classmethod
    def factory(cls, question, answer_group, value):
        question.field.check_value(question.field_parms, value)
        kwargs = {
            'answer_group':answer_group,
            'question':question,
        }
        kwargs[question.field.storage_key] = value
        
        try:
            answer = Answer.objects.get(question=question,
                answer_group=answer_group)
            setattr(answer, question.field.storage_key, value)
            answer.save()
            return answer
        except Answer.DoesNotExist:
            return Answer.objects.create(**kwargs)

    @property
    def value(self):
        """Returns the value stored in this ``Answer``.  Note that the type of
        the data returned may be a string or float depending on what is
        stored.
        """
        return getattr(self, self.question.field.storage_key)

    @property
    def display_value(self):
        """Returns a string version of the value limited to 15 characters.
        """
        value = str(self.value)
        if len(value) >= 15:
            value = value[:12] + '...'

        return value



import json, logging
from collections import OrderedDict
from functools import wraps

from django.conf import settings
from django.contrib.admin.views.decorators import staff_member_required
from django.core.urlresolvers import reverse
from django.http import HttpResponse, HttpResponseRedirect, Http404
from django.shortcuts import get_object_or_404
from django.template import Context, Template

from awl.decorators import post_required
from awl.utils import render_page
from wrench.utils import dynamic_load

from .fields import FIELDS_DICT
from .forms import SurveyForm
from .models import (EditNotAllowedException, Survey, SurveyVersion, Question,
    QuestionOrder, AnswerGroup)

logger = logging.getLogger(__name__)

# ============================================================================
# Security Decorator
# ============================================================================

def permission_hook(target):
    @wraps(target)
    def wrapper(*args, **kwargs):
        if hasattr(settings, 'DFORM_PERMISSION_HOOK'):
            fn = dynamic_load(settings.DFORM_PERMISSION_HOOK)
            fn(target.__name__, *args, **kwargs)

        # everything verified, run the view
        return target(*args, **kwargs)
    return wrapper

# ============================================================================
# AJAX Methods 
# ============================================================================

@staff_member_required
@post_required(['delta'])
def survey_delta(request, survey_version_id):
    delta = json.loads(request.POST['delta'], object_pairs_hook=OrderedDict)
    if survey_version_id == '0':
        # new survey
        survey = Survey.objects.create(name=delta['name'])
        version = survey.latest_version
    else:
        version = get_object_or_404(SurveyVersion, id=survey_version_id)

    try:
        version.replace_from_dict(delta)
    except EditNotAllowedException:
        raise Http404('Survey %s is not editable' % version.survey)
    except Question.DoesNotExist as dne:
        raise Http404('Bad question id: %s' % dne)

    # issue a 200 response
    return HttpResponse()


@staff_member_required
def survey_editor(request, survey_version_id):
    if survey_version_id == '0':
        # new survey
        survey = Survey.objects.create(name='New Survey')
        version = survey.latest_version
    else:
        version = get_object_or_404(SurveyVersion, id=survey_version_id)

    admin_link = reverse('admin:index')
    return_url = request.META.get('HTTP_REFERER', admin_link)
    save_url = reverse('dform-survey-delta', args=(version.id, ))
    data = {
        'survey_version':json.dumps(version.to_dict()),
        'save_url':save_url,
        'return_url':return_url,
    }

    return render_page(request, 'dform/edit_survey.html', data)


@staff_member_required
def new_version(request, survey_id):
    survey = get_object_or_404(Survey, id=survey_id)
    survey.new_version()

    admin_link = reverse('admin:index')
    return_url = request.META.get('HTTP_REFERER', admin_link)
    return HttpResponseRedirect(return_url)

# ============================================================================
# Form Views
# ============================================================================

@permission_hook
def sample_survey(request, survey_version_id):
    """A view for displaying a sample version of a form.  The submit mechanism
    does nothing.

    URL name reference for this view: ``dform-sample-survey``

    :param survey_version_id:
        Id of a :class:`SurveyVersion` object
    """
    version = get_object_or_404(SurveyVersion, id=survey_version_id)

    form = SurveyForm(survey_version=version)
    data = {
        'title':'Sample: %s' % version.survey.name,
        'survey_version':version,
        'form':form,
        'submit_action':'',
    }

    return render_page(request, 'dform/survey.html', data)


@permission_hook
def survey(request, survey_version_id):
    """View for submitting the answers to a survey.

    URL name reference for this view: ``dform-survey``

    """
    version = get_object_or_404(SurveyVersion, id=survey_version_id)

    try:
        template = Template(settings.DFORM_SURVEY_SUBMIT)
        context = Context({'survey_version':version})
        submit_action = template.render(context)
    except AttributeError:
        submit_action = reverse('dform-survey', args=(version.id, ))

    if request.method == 'POST':
        form = SurveyForm(request.POST, survey_version=version)
        if form.is_valid():
            form.save()
            return HttpResponseRedirect(version.on_success())
    else:
        form = SurveyForm(survey_version=version)

    data = {
        'title':version.survey.name,
        'survey_version':version,
        'form':form,
        'submit_action':submit_action,
    }

    return render_page(request, 'dform/survey.html', data)


@permission_hook
def survey_with_answers(request, survey_version_id, answer_group_id):
    """View for viewing and changing the answers to a survey that already has
    answers.

    URL name reference for this view: ``dform-survey-with-answers``
    """
    version = get_object_or_404(SurveyVersion, id=survey_version_id)
    answer_group = get_object_or_404(AnswerGroup, id=answer_group_id)

    try:
        template = Template(settings.DFORM_SURVEY_WITH_ANSWERS_SUBMIT)
        context = Context({
            'survey_version':version, 
            'answer_group':answer_group
        })
        submit_action = template.render(context)
    except AttributeError:
        submit_action = reverse('dform-survey-with-answers', args=(
            version.id, answer_group.id))

    if request.method == 'POST':
        form = SurveyForm(request.POST, survey_version=version,
            answer_group=answer_group)
        if form.is_valid():
            form.save()
            return HttpResponseRedirect(version.on_success())
    else:
        form = SurveyForm(survey_version=version, answer_group=answer_group)

    data = {
        'title':version.survey.name,
        'survey_version':version,
        'answer_group':answer_group,
        'form':form,
        'submit_action':submit_action,
    }

    return render_page(request, 'dform/survey.html', data)
